"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

from sdh.metrics.ci import app, st as store
from sdh.metrics.store.metrics import aggregate, avg
import calendar
from datetime import datetime
import math


@app.orgmetric('/total-builds', 'sum', 'builds', title='Builds')
def get_total_builds(**kwargs):
    return [len(store.get_builds())]


@app.repometric('/total-repo-builds', 'sum', 'builds', title='Builds')
def get_repo_builds(rid, **kwargs):
    return [len(store.get_repo_builds(rid))]


@app.orgmetric('/avg-builds', 'avg', 'builds', title='Builds')
def get_avg_builds(**kwargs):
    return avg([len(store.get_repo_builds(rid)) for rid in store.get_repositories()])


@app.orgmetric('/total-executions', 'sum', 'executions', title='Executions')
def get_total_executions(**kwargs):
    return aggregate(store, 'metrics:total-jobs', kwargs['begin'], kwargs['end'], kwargs['max'])


@app.orgmetric('/total-passed-executions', 'sum', 'passedexecutions', title='Executions')
def get_total_passed_executions(**kwargs):
    return aggregate(store, 'metrics:total-passed-jobs', kwargs['begin'], kwargs['end'], kwargs['max'])


@app.orgmetric('/total-failed-executions', 'sum', 'failedexecutions', title='Executions')
def get_total_failed_executions(**kwargs):
    return aggregate(store, 'metrics:total-failed-jobs', kwargs['begin'], kwargs['end'], kwargs['max'])


@app.repometric('/total-repo-executions', 'sum', 'executions', title='Executions')
def get_total_repo_executions(rid, **kwargs):
    return aggregate(store, 'metrics:total-repo-jobs:{}'.format(rid), kwargs['begin'], kwargs['end'], kwargs['max'])


@app.repometric('/avg-repo-executions', 'avg', 'executions', title='Executions')
def get_avg_repo_executions(rid, **kwargs):
    return aggregate(store, 'metrics:total-repo-jobs:{}'.format(rid), kwargs['begin'], kwargs['end'], kwargs['max'],
                     aggr=avg)


@app.repometric('/total-passed-repo-executions', 'sum', 'passedexecutions', title='Executions')
def get_total_passed_repo_executions(rid, **kwargs):
    return aggregate(store, 'metrics:total-passed-repo-jobs:{}'.format(rid), kwargs['begin'], kwargs['end'],
                     kwargs['max'])


@app.repometric('/total-failed-repo-executions', 'sum', 'failedexecutions', title='Executions')
def get_total_failed_repo_executions(rid, **kwargs):
    return aggregate(store, 'metrics:total-failed-repo-jobs:{}'.format(rid), kwargs['begin'], kwargs['end'],
                     kwargs['max'])


@app.repotbd('/repo-build-time', 'buildtime', title='Build time')
def get_repo_build_time(rid, **kwargs):
    total = store.get_repo_build_time(rid, begin=kwargs['begin'], end=kwargs['end'])
    if math.isnan(total):
        return []
    return [total]


@app.orgtbd('/avg-build-time', 'avgbuildtime', title='Build time')
def get_avg_build_time(**kwargs):
    average = avg(
        [store.get_repo_build_time(rid, begin=kwargs['begin'], end=kwargs['end']) for rid in store.get_repositories()])
    if math.isnan(average):
        return []
    return [average]


@app.orgtbd('/total-build-time', 'buildtime', title='Build time')
def get_total_build_time(**kwargs):
    total = sum(
        [store.get_repo_build_time(rid, begin=kwargs['begin'], end=kwargs['end']) for rid in store.get_repositories()])
    if math.isnan(total):
        return []

    return [total]


@app.orgtbd('/total-broken-time', 'brokentime', title='Broken time')
def get_total_broken_time(**kwargs):
    return [sum([store.get_broken_time(rid) for rid in store.get_repositories()])]


@app.repometric('/repo-broken-time', 'sum', 'brokentime', title='Broken time')
def get_repo_broken_time(rid, **kwargs):
    begin = kwargs['begin']
    if begin is None:
        begin = 0
    end = kwargs['end']
    if end is None:
        end = calendar.timegm(datetime.now().timetuple())

    return {'begin': begin, 'end': end}, [store.get_broken_time(rid, begin=begin, end=end)]


@app.repotbd('/repo-time-to-fix', 'timetofix', title='Time to fix')
def get_repo_time_to_fix(rid, **kwargs):
    begin = kwargs['begin']
    if begin is None:
        begin = 0
    end = kwargs['end']
    if end is None:
        end = calendar.timegm(datetime.now().timetuple())

    return [store.get_time_to_fix(rid, begin=begin, end=end)]


@app.orgtbd('/time-to-fix', 'timetofix', title='Time to fix')
def get_time_to_fix(**kwargs):
    begin = kwargs['begin']
    if begin is None:
        begin = 0
    end = kwargs['end']
    if end is None:
        end = calendar.timegm(datetime.now().timetuple())

    return [avg([store.get_time_to_fix(rid, begin=begin, end=end) for rid in store.get_repositories()])]



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

import calendar
from sdh.metrics.server import MetricsApp
from sdh.metrics.ci.store import CIStore
from sdh.metrics.store.metrics import store_calc
import os
import urlparse

config = os.environ.get('CONFIG', 'sdh.metrics.ci.config.DevelopmentConfig')

app = MetricsApp(__name__, config)
st = CIStore(app.config['REDIS'])
app.store = st


@st.collect('?h ci:hasBuild ?b')
def add_build((h, _, b_uri)):
    st.execute('sadd', 'frag:builds', b_uri)


@st.collect('?b ci:includesBuild ?sb')
def add_sub_build((p_uri, _, b_uri)):
    st.execute('sadd', 'frag:builds', b_uri)
    st.execute('sadd', 'frag:builds:-{}-:sub'.format(p_uri), b_uri)


@st.collect('?r scm:repositoryId ?rid')
def add_repository_id((r_uri, _, rid)):
    st.execute('hset', 'frag:repos:-{}-:'.format(r_uri), 'id', rid.toPython())
    st.execute('set', 'frag:repos:{}:'.format(rid.toPython()), r_uri)


@st.collect('?r doap:name ?rn')
def add_repository((r_uri, _, name)):
    st.execute('hset', 'frag:repos:-{}-:'.format(r_uri), 'name', name.toPython())


@st.collect('?b ci:codebase ?cb')
def link_codebase((b_uri, _, codebase)):
    codebase = codebase.toPython()
    if codebase:
        repo_name = urlparse.urlparse(codebase).path.split('/').pop(-1)
        repo_name = repo_name.replace('.git', '')
        st.execute('set', 'frag:builds:-{}-:repo'.format(b_uri), repo_name)
        st.execute('sadd', 'frag:repos:-{}-:builds'.format(repo_name), b_uri)


@st.collect('?ab ci:hasExecution ?e')
def add_execution((b_uri, _, e_uri)):
    st.execute('sadd', 'frag:builds:-{}-:jobs'.format(b_uri), e_uri)
    st.execute('sadd', 'frag:jobs:-{}-:'.format(e_uri), b_uri)


@st.collect('?e dcterms:created ?et')
def add_execution((e_uri, _, created)):
    timestamp = calendar.timegm(created.toPython().timetuple())
    st.execute('zadd', 'frag:sorted-jobs', timestamp, e_uri)
    st.execute('set', 'frag:jobs:-{}-:created'.format(e_uri), timestamp)


@st.collect('?e ci:finished ?tf')
def add_finished_execution((e_uri, _, t)):
    timestamp = calendar.timegm(t.toPython().timetuple())
    st.execute('zadd', 'frag:jobs:finished'.format(e_uri), timestamp, e_uri)
    st.execute('set', 'frag:jobs:-{}-:finished'.format(e_uri), timestamp)


@st.collect('?e oslc_auto:state ?st')
def add_execution_result((e_uri, _, state)):
    st.execute('set', 'frag:jobs:-{}-:state'.format(e_uri), state.toPython())


@st.collect('?e ci:hasResult ?jr')
def add_execution_result((e_uri, _, r_uri)):
    st.execute('set', 'frag:jobs:-{}-:result'.format(e_uri), r_uri)


@st.collect('?jr oslc_auto:verdict ?v')
def add_execution_result((jr_uri, _, verdict)):
    st.execute('set', 'frag:results:-{}-:'.format(jr_uri), verdict.toPython())


@app.calculus(triggers=['add_execution'])
def update_interval_jobs(begin, end):
    jobs = st.get_jobs(begin, end)
    store_calc(st, 'metrics:total-jobs', begin, len(jobs))

    results = [(j, st.db.get('frag:jobs:-{}-:result'.format(j))) for j in jobs]
    verdicts = filter(lambda (_, v): v is not None,
                      [(j, st.db.get('frag:results:-{}-:'.format(r))) for (j, r) in results])
    passed_jobs = [j for (j, v) in verdicts if 'passed' in v]
    failed_jobs = [j for (j, v) in verdicts if 'failed' in v]
    store_calc(st, 'metrics:total-passed-jobs', begin, len(passed_jobs))
    store_calc(st, 'metrics:total-failed-jobs', begin, len(failed_jobs))


@app.calculus(triggers=['add_execution'])
def update_interval_repo_metrics(begin, end):
    for rid in st.get_repositories():
        jobs = st.get_jobs(begin, end, rid=rid)
        store_calc(st, 'metrics:total-repo-jobs:{}'.format(rid), begin, len(jobs))

        results = [(j, st.db.get('frag:jobs:-{}-:result'.format(j))) for j in jobs]
        verdicts = filter(lambda (_, v): v is not None,
                          [(j, st.db.get('frag:results:-{}-:'.format(r))) for (j, r) in results])
        passed_jobs = [j for (j, v) in verdicts if 'passed' in v]
        failed_jobs = [j for (j, v) in verdicts if 'failed' in v]
        store_calc(st, 'metrics:total-passed-repo-jobs:{}'.format(rid), begin, len(passed_jobs))
        store_calc(st, 'metrics:total-failed-repo-jobs:{}'.format(rid), begin, len(failed_jobs))



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

__author__ = 'Fernando Serena'

from sdh.metrics.store.fragment import FragmentStore
import calendar
from datetime import datetime
import uuid


class CIStore(FragmentStore):
    def __init__(self, redis_conf):
        super(CIStore, self).__init__(redis_conf['host'], redis_conf['db'])

    def get_repositories(self):
        repo_keys = self.db.keys('frag:repos:-*-:')
        return filter(lambda x: x is not None, [self.db.hget(rk, 'id') for rk in repo_keys])

    def get_builds(self):
        return self.db.smembers('frag:builds')

    def get_repo_builds(self, rid, children=True):
        r_uri = self.db.get('frag:repos:{}:'.format(rid))
        repo_name = self.db.hget('frag:repos:-{}-:'.format(r_uri), 'name')
        try:
            builds = self.db.smembers('frag:repos:-{}-:builds'.format(repo_name))
            if children:
                builds.update(set.union(*[self.db.smembers('frag:builds:-{}-:sub'.format(b)) for b in builds]))
            return builds
        except TypeError:
            return []

    def get_jobs(self, begin=0, end=None, rid=None, bid=None, withstamps=False, limit=None, start=None, state=None):
        if end is None:
            end = calendar.timegm(datetime.utcnow().timetuple())
        jobs = self.db.zrangebyscore('frag:sorted-jobs', begin, end, withscores=withstamps,
                                     num=limit, start=start)
        if len(jobs):
            temp_key = str(uuid.uuid4())
            if rid is not None:
                self.db.sadd(temp_key, *jobs)
                repo_builds = self.get_repo_builds(rid)
                filtered_jobs = set([])
                for build in repo_builds:
                    filtered_jobs.update(self.db.sinter(temp_key, 'frag:builds:-{}-:jobs'.format(build)))
                jobs = filtered_jobs
                self.db.delete(temp_key)
            elif bid is not None:
                if withstamps:
                    self.db.zadd(temp_key, **dict(jobs))
                    ztemp_key = str(uuid.uuid4())
                    self.db.zinterstore(ztemp_key, keys=[temp_key, 'frag:builds:-{}-:jobs'.format(bid)],
                                        aggregate='MAX')
                    filtered_jobs = self.db.zrange(ztemp_key, 0, -1)
                    self.db.delete(ztemp_key)
                else:
                    self.db.sadd(temp_key, *jobs)
                    filtered_jobs = self.db.sinter(temp_key, 'frag:builds:-{}-:jobs'.format(bid))
                jobs = filtered_jobs
                self.db.delete(temp_key)

        if state is not None:
            results = [(j, self.db.get('frag:jobs:-{}-:result'.format(j))) for j in jobs]
            jobs = [j for (j, r) in results if state in self.db.get('frag:results:-{}-:'.format(r))]

        return jobs

    def get_last_passed_execution(self, bid, begin='-inf', end='+inf'):
        jobs = self.db.smembers('frag:builds:-{}-:jobs'.format(bid))
        results = [(j, self.db.get('frag:jobs:-{}-:result'.format(j))) for j in jobs]
        jobs = [j for (j, r) in results if 'pass' in self.db.get('frag:results:-{}-:'.format(r))]
        temp_key = str(uuid.uuid4())
        ztemp_key = str(uuid.uuid4())
        stemp_key = str(uuid.uuid4())
        sorted_key = 'frag:sorted-jobs'
        if begin != '-inf' or end != '+inf':
            filtered_jobs = self.db.zrangebyscore(sorted_key, begin, end, withscores=True)
            if filtered_jobs:
                self.db.zadd(stemp_key, **dict(filtered_jobs))
            sorted_key = stemp_key
        self.db.sadd(temp_key, *jobs)
        self.db.zinterstore(ztemp_key, keys=[temp_key, sorted_key], aggregate='MAX')
        self.db.delete(temp_key)
        self.db.delete(stemp_key)
        result = self.db.zrevrangebyscore(ztemp_key, '+inf', '-inf', start=0, num=1)
        self.db.delete(ztemp_key)

        try:
            return result.pop()
        except IndexError:
            return None

    def get_job_build_time(self, e_uri):
        finished = self.db.get('frag:jobs:-{}-:finished'.format(e_uri))
        created = self.db.get('frag:jobs:-{}-:created'.format(e_uri))
        try:
            return int(finished) - int(created)
        except TypeError:
            return float('NaN')

    def get_repo_build_time(self, rid, begin='-inf', end='+inf'):
        repo_builds = self.get_repo_builds(rid, children=False)
        build_time = 0
        for b_uri in repo_builds:
            last_exec = self.get_last_passed_execution(b_uri, begin, end)
            if last_exec is not None:
                build_time += self.get_job_build_time(last_exec)
        return build_time

    def get_broken_time(self, rid, begin=0, end=None):
        if end is None:
            end = calendar.timegm(datetime.utcnow().timetuple())
        builds = self.get_repo_builds(rid, children=False)
        repo_broken_time = 0
        for b in builds:
            jobs = self.get_jobs(begin, end, bid=b, withstamps=True)
            last_verdict = 'passed'
            build_broken_since = 0
            interval_broken_time = 0
            build_broken_time = 0
            for job in jobs:
                result = self.db.get('frag:jobs:-{}-:result'.format(job))
                verdict = self.db.get('frag:results:-{}-:'.format(result))
                if last_verdict == 'passed' and last_verdict not in verdict:
                    build_broken_since = int(self.db.get('frag:jobs:-{}-:finished'.format(job)))
                elif last_verdict == 'failed' and last_verdict in verdict:
                    next_broken_timestamp = int(self.db.get('frag:jobs:-{}-:finished'.format(job)))
                    interval_broken_time = next_broken_timestamp - build_broken_since
                elif last_verdict == 'failed' and last_verdict not in verdict:
                    next_pass_timestamp = int(self.db.get('frag:jobs:-{}-:finished'.format(job)))
                    build_broken_time += (next_pass_timestamp - build_broken_since)
                    interval_broken_time = 0
                last_verdict = verdict.split('#')[1]
            if last_verdict == 'failed':
                build_broken_time += interval_broken_time
            repo_broken_time += build_broken_time
        return repo_broken_time

    def get_time_to_fix(self, rid, begin=0, end=None):
        if end is None:
            end = calendar.timegm(datetime.utcnow().timetuple())
        builds = self.get_repo_builds(rid, children=False)
        repo_time_to_fix = 0
        intervals = 0
        for b in builds:
            jobs = self.get_jobs(begin, end, bid=b, withstamps=True)
            last_verdict = 'passed'
            build_broken_since = 0
            for job in jobs:
                result = self.db.get('frag:jobs:-{}-:result'.format(job))
                verdict = self.db.get('frag:results:-{}-:'.format(result))
                if last_verdict == 'passed' and last_verdict not in verdict:
                    build_broken_since = int(self.db.get('frag:jobs:-{}-:finished'.format(job)))
                elif last_verdict == 'failed' and last_verdict not in verdict:
                    intervals += 1
                    next_pass_timestamp = int(self.db.get('frag:jobs:-{}-:finished'.format(job)))
                    repo_time_to_fix += (next_pass_timestamp - build_broken_since)
                last_verdict = verdict.split('#')[1]
        if intervals:
            repo_time_to_fix /= float(intervals)
        return repo_time_to_fix
